# -*- coding: utf-8 -*-


import sys
from logging import Formatter as FormatterBase
from logging import StreamHandler, captureWarnings, getLogger
from logging.handlers import WatchedFileHandler

from .serializer import serialize
from .utils import current_task


class Formatter(FormatterBase):
    def __init__(self, wakaq):
        self.wakaq = wakaq
        super().__init__(wakaq.log_format)

    def format(self, record):
        task = current_task.get()
        if task is not None:
            task, payload = task[0], task[1]
            self._fmt = self.wakaq.task_log_format
            self._style._fmt = self.wakaq.task_log_format
            record.__dict__.update(task=task.name)
            record.__dict__.update(task_args=serialize(payload["args"]))
            record.__dict__.update(task_kwargs=serialize(payload["kwargs"]))
            record.__dict__.update(task_retry=serialize(payload.get("retry")))
        else:
            self._fmt = self.wakaq.log_format
            self._style._fmt = self.wakaq.log_format
            record.__dict__.setdefault("task", None)
            record.__dict__.setdefault("task_args", None)
            record.__dict__.setdefault("task_kwargs", None)
            record.__dict__.setdefault("task_retry", None)
        return super().format(record)


def setup_logging(wakaq, is_child=None, is_scheduler=None):
    logger = getLogger("wakaq")

    for handler in logger.handlers:
        logger.removeHandler(handler)

    log_file = wakaq.scheduler_log_file if is_scheduler else wakaq.worker_log_file
    log_level = wakaq.scheduler_log_level if is_scheduler else wakaq.worker_log_level

    logger.setLevel(log_level)
    captureWarnings(True)

    out = sys.stdout if is_child or not log_file else log_file
    options = {}
    if not is_child and log_file:
        options["encoding"] = "utf8"
    handler = (StreamHandler if is_child or not log_file else WatchedFileHandler)(out, **options)
    handler.setLevel(log_level)

    formatter = Formatter(wakaq)
    handler.setFormatter(formatter)

    logger.addHandler(handler)



# -*- coding: utf-8 -*-


import ast
import calendar
import operator
import os
import sys
from datetime import datetime, timedelta
from importlib import import_module
from typing import Union

from .serializer import deserialize


def import_app(app):
    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.insert(0, cwd)
    module_path, class_name = app.rsplit(".", 1)
    module = import_module(module_path)
    wakaq = getattr(module, class_name)
    from . import WakaQ

    if not isinstance(wakaq, WakaQ):
        raise Exception(f"Invalid app path: {app}. App must point to a WakaQ instance.")
    return wakaq


def inspect(app):
    queues = {}
    for queue in app.queues:
        queues[queue.name] = {
            "name": queue.name,
            "priority": queue.priority,
            "broker_key": queue.broker_key,
            "broker_eta_key": queue.broker_eta_key,
            "pending_tasks": num_pending_tasks_in_queue(app, queue),
            "pending_eta_tasks": num_pending_eta_tasks_in_queue(app, queue),
        }
    return {
        "queues": queues,
        "workers": num_workers_connected(app),
    }


def pending_tasks_in_queue(app, queue=None, queue_name: str = None, limit: int = None) -> list:
    if not queue:
        if queue_name is None:
            return []
        queue = app.queues_by_name.get(queue_name)
        if not queue:
            return []

    if not limit:
        limit = 0

    tasks = app.broker.lrange(queue.broker_key, 0, limit - 1)
    return [deserialize(task) for task in tasks]


def pending_eta_tasks_in_queue(
    app,
    queue=None,
    queue_name: str = None,
    before: Union[datetime, timedelta, int] = None,
    limit: int = None,
) -> list:
    if not queue:
        if queue_name is None:
            return []
        queue = app.queues_by_name.get(queue_name)
        if not queue:
            return []
    params = []
    if before:
        cmd = "ZRANGEBYSCORE"
        if isinstance(before, timedelta):
            before = datetime.utcnow() + before
        if isinstance(before, datetime):
            before = calendar.timegm(before.utctimetuple())
        params.extend([0, before])
        params.append("WITHSCORES")
        if limit:
            params.extend(["LIMIT", 0, limit])
    else:
        cmd = "ZRANGE"
        if not limit:
            limit = 0
        params.extend([0, limit - 1])
        params.append("WITHSCORES")
    tasks = app.broker.execute_command(cmd, queue.broker_eta_key, *params)
    payloads = []
    for n in range(0, len(tasks), 2):
        payload = deserialize(tasks[n])
        payload["eta"] = datetime.utcfromtimestamp(int(tasks[n + 1]))
        payloads.append(payload)
    return payloads


def num_pending_tasks_in_queue(app, queue=None, queue_name: str = None) -> int:
    if not queue:
        if queue_name is None:
            return 0
        queue = app.queues_by_name.get(queue_name)
        if not queue:
            return 0
    return app.broker.llen(queue.broker_key)


def num_pending_eta_tasks_in_queue(app, queue=None, queue_name: str = None) -> int:
    if not queue:
        if queue_name is None:
            return 0
        queue = app.queues_by_name.get(queue_name)
        if not queue:
            return 0
    return app.broker.zcount(queue.broker_eta_key, "-inf", "+inf")


def num_workers_connected(app) -> int:
    return app.broker.pubsub_numsub(app.broadcast_key)[0][1]


def purge_queue(app, queue_name: str):
    if queue_name is None:
        return
    queue = app.queues_by_name.get(queue_name)
    if not queue:
        return
    app.broker.delete(queue.broker_key)


def purge_eta_queue(app, queue_name: str):
    if queue_name is None:
        return
    queue = app.queues_by_name.get(queue_name)
    if not queue:
        return
    app.broker.delete(queue.broker_eta_key)


def kill(pid, signum):
    try:
        os.kill(pid, signum)
    except IOError:
        pass


def read_fd(fd):
    try:
        return os.read(fd, 64000).decode("utf8")
    except OSError:
        return ""


def write_fd_or_raise(fd, s):
    os.write(fd, s.encode("utf8"))


def write_fd(fd, s):
    try:
        write_fd_or_raise(fd, s)
    except:
        pass


def close_fd(fd):
    try:
        os.close(fd)
    except:
        pass


def flush_fh(fh):
    try:
        fh.flush()
    except:
        pass


_operations = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.FloorDiv: operator.floordiv,
    ast.Pow: operator.pow,
}


def _safe_eval(node, variables, functions):
    if isinstance(node, ast.Num):
        return node.n
    elif isinstance(node, ast.Name):
        try:
            return variables[node.id]
        except KeyError:
            raise Exception(f"Unknown variable: {node.id}")
    elif isinstance(node, ast.BinOp):
        try:
            op = _operations[node.op.__class__]
        except KeyError:
            raise Exception(f"Unknown operation: {node.op.__class__}")
        left = _safe_eval(node.left, variables, functions)
        right = _safe_eval(node.right, variables, functions)
        if isinstance(node.op, ast.Pow):
            assert right < 100
        return op(left, right)
    elif isinstance(node, ast.Call):
        assert not node.keywords and not node.starargs and not node.kwargs
        assert isinstance(node.func, ast.Name), "Unsafe function derivation"
        try:
            func = functions[node.func.id]
        except KeyError:
            raise Exception(f"Unknown function: {node.func.id}")
        args = [_safe_eval(arg, variables, functions) for arg in node.args]
        return func(*args)
    assert False, "Unsafe operation"


def safe_eval(expr, variables={}, functions={}):
    node = ast.parse(expr, "<string>", "eval").body
    return _safe_eval(node, variables, functions)


class Context:
    __slots__ = ["value"]

    def __init__(self):
        self.value = None

    def get(self):
        return self.value

    def set(self, val):
        self.value = val


current_task = Context()


def exception_in_chain(e, exception_type):
    while e.__context__ is not None:
        if isinstance(e.__context__, exception_type):
            return True
        e = e.__context__
    return False



# -*- coding: utf-8 -*-


import re
from datetime import timedelta


class Queue:
    __slots__ = [
        "name",
        "priority",
        "prefix",
        "soft_timeout",
        "hard_timeout",
        "max_retries",
    ]

    def __init__(self, name=None, priority=-1, prefix=None, soft_timeout=None, hard_timeout=None, max_retries=None):
        self.prefix = re.sub(r"[^a-zA-Z0-9_.-]", "", prefix or "wakaq")
        self.name = re.sub(r"[^a-zA-Z0-9_.-]", "", name)

        try:
            self.priority = int(priority)
        except:
            raise Exception(f"Invalid queue priority: {priority}")

        self.soft_timeout = soft_timeout.total_seconds() if isinstance(soft_timeout, timedelta) else soft_timeout
        self.hard_timeout = hard_timeout.total_seconds() if isinstance(hard_timeout, timedelta) else hard_timeout

        if self.soft_timeout and self.hard_timeout and self.hard_timeout <= self.soft_timeout:
            raise Exception(
                f"Queue hard timeout ({self.hard_timeout}) can not be less than or equal to soft timeout ({self.soft_timeout})."
            )

        if max_retries:
            try:
                self.max_retries = int(max_retries)
            except:
                raise Exception(f"Invalid queue max retries: {max_retries}")
        else:
            self.max_retries = None

    @classmethod
    def create(cls, obj, queues_by_name=None):
        if isinstance(obj, cls):
            if queues_by_name is not None and obj.name not in queues_by_name:
                raise Exception(f"Unknown queue: {obj.name}")
            return obj
        elif isinstance(obj, (list, tuple)) and len(obj) == 2:
            if isinstance(obj[0], int):
                if queues_by_name is not None and obj[1] not in queues_by_name:
                    raise Exception(f"Unknown queue: {obj[1]}")
                return cls(priority=obj[0], name=obj[1])
            else:
                if queues_by_name is not None and obj[0] not in queues_by_name:
                    raise Exception(f"Unknown queue: {obj[0]}")
                return cls(name=obj[0], priority=obj[1])
        else:
            if queues_by_name is not None and obj not in queues_by_name:
                raise Exception(f"Unknown queue: {obj}")
            return cls(name=obj)

    @property
    def broker_key(self):
        return f"{self.prefix}:{self.name}"

    @property
    def broker_eta_key(self):
        return f"{self.prefix}:eta:{self.name}"



__title__ = "WakaQ"
__description__ = "Background task queue for Python backed by Redis, a minimal Celery."
__url__ = "https://github.com/wakatime/wakaq"
__version_info__ = ("2", "1", "9")
__version__ = ".".join(__version_info__)
__author__ = "Alan Hamlett"
__author_email__ = "alan.hamlett@gmail.com"
__license__ = "BSD"
__copyright__ = "Copyright 2022 Alan Hamlett"



# -*- coding: utf-8 -*-


import calendar
import logging
import multiprocessing
from datetime import datetime, timedelta

import redis

from .queue import Queue
from .scheduler import CronTask
from .serializer import serialize
from .task import Task
from .utils import safe_eval

__all__ = [
    "WakaQ",
    "Queue",
    "CronTask",
]


class WakaQ:
    queues = []
    soft_timeout = None
    hard_timeout = None
    concurrency = 0
    schedules = []
    exclude_queues = []
    max_retries = None
    wait_timeout = None
    max_mem_percent = None
    max_tasks_per_worker = None
    worker_log_file = None
    scheduler_log_file = None
    worker_log_level = None
    scheduler_log_level = None

    after_worker_started_callback = None
    before_task_started_callback = None
    after_task_finished_callback = None
    wrap_tasks_function = None

    broadcast_key = "wakaq-broadcast"
    log_format = "[%(asctime)s] %(levelname)s: %(message)s"
    task_log_format = "[%(asctime)s] %(levelname)s in %(task)s: %(message)s"

    def __init__(
        self,
        queues=[],
        schedules=[],
        host="localhost",
        port=6379,
        username=None,
        password=None,
        db=0,
        concurrency=0,
        exclude_queues=[],
        max_retries=None,
        soft_timeout=None,
        hard_timeout=None,
        max_mem_percent=None,
        max_tasks_per_worker=None,
        worker_log_file=None,
        scheduler_log_file=None,
        worker_log_level=None,
        scheduler_log_level=None,
        socket_timeout=15,
        socket_connect_timeout=15,
        health_check_interval=30,
        wait_timeout=1,
    ):
        self.queues = [Queue.create(x) for x in queues]
        if len(self.queues) == 0:
            raise Exception("Missing queues.")
        lowest_priority = max(self.queues, key=lambda q: q.priority)
        self.queues = list(map(lambda q: self._default_priority(q, lowest_priority.priority), self.queues))
        self.queues.sort(key=lambda q: q.priority)
        self.queues_by_name = dict([(x.name, x) for x in self.queues])
        self.queues_by_key = dict([(x.broker_key, x) for x in self.queues])
        self.exclude_queues = self._validate_queue_names(exclude_queues)
        self.max_retries = int(max_retries or 0)
        self.broker_keys = [x.broker_key for x in self.queues if x.name not in self.exclude_queues]
        self.schedules = [CronTask.create(x) for x in schedules]
        self.concurrency = self._format_concurrency(concurrency)
        self.soft_timeout = soft_timeout.total_seconds() if isinstance(soft_timeout, timedelta) else soft_timeout
        self.hard_timeout = hard_timeout.total_seconds() if isinstance(hard_timeout, timedelta) else hard_timeout
        self.wait_timeout = wait_timeout.total_seconds() if isinstance(wait_timeout, timedelta) else wait_timeout

        if self.soft_timeout and self.soft_timeout <= wait_timeout:
            raise Exception(
                f"Soft timeout ({self.soft_timeout}) can not be less than or equal to wait timeout ({self.wait_timeout})."
            )
        if self.hard_timeout and self.hard_timeout <= self.wait_timeout:
            raise Exception(
                f"Hard timeout ({self.hard_timeout}) can not be less than or equal to wait timeout ({self.wait_timeout})."
            )
        if self.soft_timeout and self.hard_timeout and self.hard_timeout <= self.soft_timeout:
            raise Exception(
                f"Hard timeout ({self.hard_timeout}) can not be less than or equal to soft timeout ({self.soft_timeout})."
            )

        if max_mem_percent:
            self.max_mem_percent = int(max_mem_percent)
            if self.max_mem_percent < 1 or self.max_mem_percent > 99:
                raise Exception(f"Max memory percent must be between 1 and 99: {self.max_mem_percent}")
        else:
            self.max_mem_percent = None

        self.max_tasks_per_worker = abs(int(max_tasks_per_worker)) if max_tasks_per_worker else None
        self.worker_log_file = worker_log_file if isinstance(worker_log_file, str) else None
        self.scheduler_log_file = scheduler_log_file if isinstance(scheduler_log_file, str) else None
        self.worker_log_level = worker_log_level if isinstance(worker_log_level, int) else logging.INFO
        self.scheduler_log_level = scheduler_log_level if isinstance(scheduler_log_level, int) else logging.DEBUG

        self.tasks = {}
        self.broker = redis.Redis(
            host=host,
            port=port,
            username=username,
            password=password,
            db=db,
            charset="utf-8",
            decode_responses=True,
            health_check_interval=health_check_interval,
            socket_timeout=socket_timeout,
            socket_connect_timeout=socket_connect_timeout,
        )

    def task(self, fn=None, queue=None, max_retries=None, soft_timeout=None, hard_timeout=None):
        def wrap(f):
            t = Task(
                fn=f,
                wakaq=self,
                queue=queue,
                max_retries=max_retries,
                soft_timeout=soft_timeout,
                hard_timeout=hard_timeout,
            )
            if t.name in self.tasks:
                raise Exception(f"Duplicate task name: {t.name}")
            self.tasks[t.name] = t
            return t.fn

        return wrap(fn) if fn else wrap

    def after_worker_started(self, callback):
        self.after_worker_started_callback = callback
        return callback

    def before_task_started(self, callback):
        self.before_task_started_callback = callback
        return callback

    def after_task_finished(self, callback):
        self.after_task_finished_callback = callback
        return callback

    def wrap_tasks_with(self, callback):
        self.wrap_tasks_function = callback
        return callback

    def _validate_queue_names(self, queue_names: list) -> list:
        try:
            queue_names = [x for x in queue_names]
        except:
            return []
        for queue_name in queue_names:
            if queue_name not in self.queues_by_name:
                raise Exception(f"Invalid queue: {queue_name}")
        return queue_names

    def _enqueue_at_front(self, task_name: str, queue: str, args: list, kwargs: dict):
        queue = self._queue_or_default(queue)
        payload = serialize(
            {
                "name": task_name,
                "args": args,
                "kwargs": kwargs,
            }
        )
        self.broker.lpush(queue.broker_key, payload)

    def _enqueue_at_end(self, task_name: str, queue: str, args: list, kwargs: dict, retry=0):
        queue = self._queue_or_default(queue)
        payload = serialize(
            {
                "name": task_name,
                "args": args,
                "kwargs": kwargs,
                "retry": retry,
            }
        )
        self.broker.rpush(queue.broker_key, payload)

    def _enqueue_with_eta(self, task_name: str, queue: str, args: list, kwargs: dict, eta):
        queue = self._queue_or_default(queue)
        payload = serialize(
            {
                "name": task_name,
                "args": args,
                "kwargs": kwargs,
            }
        )
        if isinstance(eta, timedelta):
            eta = datetime.utcnow() + eta
        timestamp = calendar.timegm(eta.utctimetuple())
        self.broker.zadd(queue.broker_eta_key, {payload: timestamp}, nx=True)

    def _broadcast(self, task_name: str, args: list, kwargs: dict):
        payload = serialize(
            {
                "name": task_name,
                "args": args,
                "kwargs": kwargs,
            }
        )
        return self.broker.publish(self.broadcast_key, payload)

    def _queue_or_default(self, queue_name: str):
        if queue_name:
            return Queue.create(queue_name, queues_by_name=self.queues_by_name)

        # return lowest priority queue by default
        return self.queues[-1]

    def _default_priority(self, queue, lowest_priority):
        if queue.priority < 0:
            queue.priority = lowest_priority + 1
        return queue

    def _format_concurrency(self, concurrency):
        if not concurrency:
            return 0
        try:
            return safe_eval(str(concurrency), {"cores": multiprocessing.cpu_count()})
        except Exception as e:
            raise Exception(f"Error parsing concurrency: {e}")
